""" 

Optimized Gaussian Kernel generator script by The Sandvich Maker

MIT License

Copyright (c) 2018 Daniel Cornelisse

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import sys


def pascalTriangle(n):
    """ Calculates a Pascal Triangle with n rows """

    n = n + 1
    triangle = [[0] * n for _ in range(n)]
    triangle[0][0] = 1
    triangle[1][0] = triangle[1][1] = 1

    for i in range(2, n):
        for j in range(0, n):
            triangle[i][j] = triangle[i - 1][j - 1] + triangle[i - 1][j]

    return triangle


def generateGaussian(taps):
    """ Generates a gaussian convolution kernel the kernel 
        has been chopped in half since it is symmetrical """

    if taps % 2 == 0:
        taps = taps + 1
        print 'Number of taps has to be an odd number! ' + \
            'Moved number of taps to ' + str(taps)

    n = taps + 3
    p = pascalTriangle(n)[-1][1:-2]
    p = p[len(p) / 2:]

    divider = pow(2, n) - n - n - 2
    p = [x / float(divider) for x in p]

    p = [p, [0] * len(p)]

    for i in range(0, len(p[1])):
        p[1][i] = float(i)

    return p


def optimizedGaussian(taps):
    """ Generates a gaussian kernel with offsets and 
        weights calculated to exploit GPU lerping """

    g = generateGaussian(taps)[0]

    length = (len(g) - 1) / 2 + 1
    gW = [g[0]] * length
    gO = [0.0] * length

    for i in range(0, len(gW) - 1):
        j = i * 2 + 1
        gW[i + 1] = g[j] + g[j + 1]

    for i in range(0, len(gO) - 1):
        j = i * 2 + 1
        gO[i + 1] = (j * g[j] + (j + 1) * g[j + 1]) / gW[i + 1]

    return [gW, gO]


# defaults
Quiet = False

Taps = 17
Optimize = True

Prefix = 'gaussian'
OutFile = 'gaussianWeights.fxh'
WriteMode = 'w'
FinalWrite = True

if __name__ == '__main__':
    # handle arguments
    if len(sys.argv) > 1:
        for i in range(1, len(sys.argv)):
            arg = sys.argv[i]
            if arg == '--taps':
                Taps = int(sys.argv[i + 1])
            if arg == '--append':
                WriteMode = 'a'
            if arg == '--nowrite':
                WriteMode = None
            if arg == '--quiet':
                Quiet = True
            if arg == '--prefix':
                Prefix = sys.argv[i + 1]
            if arg == '--unoptimized':
                Optimize = False
            if arg == '--out':
                OutFile = sys.argv[i + 1]
            if arg == '--notfinal':
                FinalWrite = False

    # generate gaussian
    g = None
    if Optimize:
        g = optimizedGaussian(Taps)
    else:
        g = generateGaussian(Taps)

    # output result
    num = len(g[0])

    outS = """
/* This code was automatically generated by gaussianGenerator.py
   Number of filter taps per pass: {taps}
   Number of texture samples per pass: {samples} */
   \n""".format(taps=Taps, samples=num * 2 - 1)

    if WriteMode == 'w':
        outS = outS + """\
#ifndef GAUSSIAN_WEIGHTS_H
#define GAUSSIAN_WEIGHTS_H\n"""
    else:
        outS = outS + """\
#ifndef GAUSSIAN_ANAMORPHIC
#define GAUSSIAN_ANAMORPHIC 1
#endif\n"""

    outS = outS + 'static const float {prefix}Weights[{n}] = {{ '.format(
            prefix=Prefix, n=num)
    for i in range(0, num):
        outS = outS + str(g[0][i])
        if i != num - 1:
            outS = outS + ', '
    outS = outS + ' };\n'

    outS = outS + 'static const float {prefix}Offsets[{n}] = {{ '.format(
            prefix=Prefix, n=num)
    for i in range(0, num):
        outS = outS + str(g[1][i])
        if i != num - 1:
            outS = outS + ', '
    outS = outS + ' };\n'

    outS = outS + 'static const int {prefix}LoopLength = {n};\n'.format(
            prefix=Prefix, n=num)

    if FinalWrite:
        outS = outS + '#endif\n'

    if not Quiet:
        print outS

    if WriteMode is not None:
        f = open(OutFile, WriteMode)
        f.write(outS)
        f.close()

    print """\
generated {tap} tap gaussian
optimized: {optimized}
""".format(tap=Taps, optimized=Optimize)
